/*
 *  core_es.cpp
 *  $Id: core_es.c 20 2011-11-29 12:32:47Z ijuranek $
 *
 *  Author
 *  Roman Juranek <ijuranek@fit.vutbr.cz>
 *
 *  Graph@FIT
 *  Department of Computer Graphics and Multimedia
 *  Faculty of Information Technology
 *  Brno University of Technology
 *
 */


#include "core_es.h"
#include <assert.h>

// TODO eval_stage_es

static inline int eval_classifier_es(
        const char * img, unsigned width_step,
        TClassifier * c,
        unsigned begin, unsigned end,
        float * response)
{
    for (TStage * s = c->stage + begin; s != c->stage + end; ++s)
    {
        *response += eval_stage_es(img, width_step, s);
        if (*response < stg->theta_b)
        {
            return 0;
        }
    }
    return 1;
}


int scan_image_intensity_es(
        const char * image_data, unsigned width_step, int w, int h,
        TClassifier * c,
        ScanParams * sp,
        Detection * first, Detection * last,
        int * hist)
{
    assert(image_data != 0);
    assert(width_step >= w);
    assert(c != 0);
    assert(sp != 0);
    assert(first != 0 && last >= first);

    if (c->tp != LBP) // FSZ_2x2
        return 0;

    Detection * det = first;
    unsigned start_stage = (sp->start_stage >= 0) ? sp->start_stage : 0;
    unsigned end_stage = (sp->end_stage >= 0) ? sp->end_stage : c->stage_count;

    for (unsigned y = 1; y < h-c->height - 1; y++)
    {
        // ptr to map;
        float * map = (float*)(sp->response_map->imageData + y * sp->response_map->widthStep);
        unsigned row_offset = y * width_step;
        for (unsigned x = 1; x < w-c->width-1; x++)
        {
            response = 0.0f;
            bool scan = true;
            if (sp->response_map)
            {
                if (map[x] > -1E30)
                {
                    response = map;
                }
                else
                {
                    scan = false;
                }
            }

            if (!scan) continue;
            
            int d = eval(row_offset + x, width_step,...);

            if (d && (response > c->threshold))
            {
                Detection tmp = {x, y, c->width, c->height, response, 0.0f};
                *det = tmp;
                if (det == last) return det - first;
            }

        }
    }

    return det - first;    
}
